import os
import copy
import numpy as np
import pandas as pd
pd.options.display.max_colwidth = 100
from pyemu.pst.pst_utils import SFMT,IFMT,FFMT,pst_config
from pyemu.utils.helpers import run
from ..pyemu_warnings import PyemuWarning
PP_FMT = {"name": SFMT, "x": FFMT, "y": FFMT, "zone": IFMT, "tpl": SFMT,
          "parval1": FFMT}
PP_NAMES = ["name","x","y","zone","parval1"]


def setup_pilotpoints_grid(ml=None,sr=None,ibound=None,prefix_dict=None,
                           every_n_cell=4,
                           use_ibound_zones=False,
                           pp_dir='.',tpl_dir='.',
                           shapename="pp.shp"):
    """setup grid-based pilot points.  Uses the ibound to determine
    where to set pilot points. pilot points are given generic ``pp_``
    names unless ``prefix_dict`` is passed.  Write template files and a
    shapefile as well...hopefully this is useful to someone...

    Parameters
    ----------
    ml : flopy.mbase
        a flopy mbase dervied type.  If None, sr must not be None.
    sr : flopy.utils.reference.SpatialReference
        a spatial reference use to locate the model grid in space.  If None,
        ml must not be None.  Default is None
    ibound : numpy.ndarray
        the modflow ibound integer array.  Used to set pilot points only in active areas (!=0).
        If None and ml is None, then pilot points are set in all rows and columns according to
        every_n_cell.  Default is None.
    prefix_dict : dict
        a dictionary of layer index, pilot point parameter prefixes.
        Example : ``{0:["hk_"],1:["vka_"]}`` would setup pilot points with
        the prefix ``"hk"`` for model layers 1 and ``"vka"`` for model layer 2 (zero based).
        If None, a generic set of pilot points with the "pp" prefix are setup
        for a generic nrowXncol grid. Default is None.
    use_ibound_zones : bool
        a flag to use the greater-than-zero values in the ibound as pilot point zones.  If False,ibound
        values greater than zero are treated as a single zone.  Default is False.
    pp_dir : str
        directory to write pilot point files to.  Default is '.'
    tpl_dir : str
        directory to write pilot point template file to.  Default is '.'
    shapename : str
        name of shapefile to write that contains pilot point information. Default is "pp.shp"

    Returns
    -------
    pp_df : pandas.DataFrame
        a dataframe summarizing pilot point information (same information
        written to shapename
    Example
    -------
    ``>>>import flopy``

    ``>>>from pyemu.utils import setup_pilotpoints_grid``

    ``>>>m = flopy.modflow.Modfow.load("mymodel.nam")``

    ``>>>setup_pilotpoints_grid(m,prefix_dict={0:['hk_'],1:['vka_']},``

    ``>>>                       every_n_cell=3,shapename='layer1_pp.shp')``
    """

    import flopy

    if ml is not None:
        assert isinstance(ml,flopy.modflow.Modflow)
        sr = ml.sr
        if ibound is None:
            ibound = ml.bas6.ibound.array
    else:
        assert sr is not None,"if 'ml' is not passed, 'sr' must be passed"
        if ibound is None:
            print("ibound not passed, using array of ones")
            ibound = {k:np.ones((sr.nrow,sr.ncol)) for k in prefix_dict.keys()}
        #assert ibound is not None,"if 'ml' is not pass, 'ibound' must be passed"

    if isinstance(ibound, np.ndarray):
        assert np.ndim(ibound) == 2 or np.ndim(ibound) == 3, \
            "ibound needs to be either 3d np.ndarry or k_dict of 2d arrays. " \
            "Array of {0} dimensions passed".format(np.ndim(ibound))
        if np.ndim(ibound) == 2:
            ibound = {0: ibound}
        else:
            ibound = {k: arr for k, arr in enumerate(ibound)}


    try:
        xcentergrid = sr.xcentergrid
        ycentergrid = sr.ycentergrid
    except Exception as e:
        raise Exception("error getting xcentergrid and/or ycentergrid from 'sr':{0}".format(str(e)))
    start = int(float(every_n_cell) / 2.0)

    #build a generic prefix_dict
    if prefix_dict is None:
        prefix_dict = {k:["pp_{0:02d}_".format(k)] for k in range(ml.nlay)}

    #check prefix_dict
    keys = list(prefix_dict.keys())
    keys.sort()
    #for k, prefix in prefix_dict.items():
    for k in keys:
        prefix = prefix_dict[k]
        if not isinstance(prefix,list):
            prefix_dict[k] = [prefix]
        if np.all([isinstance(v, dict) for v in ibound.values()]):
            for p in prefix_dict[k]:
                if np.any([p.startswith(key) for key in ibound.keys()]):
                    ib_sel = next(key for key in ibound.keys() if p.startswith(key))
                else:
                    ib_sel = 'general_zn'
                assert k < len(ibound[ib_sel]), "layer index {0} > nlay {1}".format(k, len(ibound[ib_sel]))
        else:
            assert k < len(ibound),"layer index {0} > nlay {1}".format(k,len(ibound))


    #try:
        #ibound = ml.bas6.ibound.array
    #except Exception as e:
    #    raise Exception("error getting model.bas6.ibound:{0}".format(str(e)))
    par_info = []
    pp_files,tpl_files = [],[]
    pp_names = copy.copy(PP_NAMES)
    pp_names.extend(["k","i","j"])

    if not np.all([isinstance(v, dict) for v in ibound.values()]):
        ibound = {"general_zn": ibound}
    par_keys = list(ibound.keys())
    par_keys.sort()
    for par in par_keys:
        for k in range(len(ibound[par])):
            pp_df = None
            ib = ibound[par][k]
            assert ib.shape == xcentergrid.shape,"ib.shape != xcentergrid.shape for k {0}".\
                format(k)
            pp_count = 0
            #skip this layer if not in prefix_dict
            if k not in prefix_dict.keys():
                continue
            #cycle through rows and cols
            for i in range(start,ib.shape[0]-start,every_n_cell):
                for j in range(start,ib.shape[1]-start,every_n_cell):
                    # skip if this is an inactive cell
                    if ib[i,j] == 0:
                        continue

                    # get the attributes we need
                    x = xcentergrid[i,j]
                    y = ycentergrid[i,j]
                    name = "pp_{0:04d}".format(pp_count)
                    parval1 = 1.0

                    #decide what to use as the zone
                    zone = 1
                    if use_ibound_zones:
                        zone = ib[i,j]
                    #stick this pilot point into a dataframe container

                    if pp_df is None:
                        data = {"name": name, "x": x, "y": y, "zone": zone,
                                "parval1": parval1, "k":k, "i":i, "j":j}
                        pp_df = pd.DataFrame(data=data,index=[0],columns=pp_names)
                    else:
                        data = [name, x, y, zone, parval1, k, i, j]
                        pp_df.loc[pp_count,:] = data
                    pp_count += 1
            #if we found some acceptable locs...
            if pp_df is not None:
                for prefix in prefix_dict[k]:
                    # if parameter prefix relates to current zone definition
                    if prefix.startswith(par) or (
                            ~np.any([prefix.startswith(p) for p in ibound.keys()]) and par == "general_zn"):
                        base_filename = prefix+"pp.dat"
                        pp_filename = os.path.join(pp_dir, base_filename)
                        # write the base pilot point file
                        write_pp_file(pp_filename, pp_df)

                        tpl_filename = os.path.join(tpl_dir, base_filename + ".tpl")
                        #write the tpl file
                        pilot_points_to_tpl(pp_df, tpl_filename,
                                            name_prefix=prefix)
                        pp_df.loc[:,"tpl_filename"] = tpl_filename
                        pp_df.loc[:,"pp_filename"] = pp_filename
                        pp_df.loc[:,"pargp"] = prefix
                        #save the parameter names and parval1s for later
                        par_info.append(pp_df.copy())
                        #save the pp_filename and tpl_filename for later
                        pp_files.append(pp_filename)
                        tpl_files.append(tpl_filename)


    par_info = pd.concat(par_info)
    for field in ["k","i","j"]:
        par_info.loc[:,field] = par_info.loc[:,field].apply(np.int)
    for key,default in pst_config["par_defaults"].items():
        if key in par_info.columns:
            continue
        par_info.loc[:,key] = default

    if shapename is not None:
        try:
            import shapefile
        except Exception as e:
            print("error importing shapefile, try pip install pyshp...{0}"\
                  .format(str(e)))
            return par_info
        try:
            shp = shapefile.Writer(target=shapename,shapeType=shapefile.POINT)
        except:
            shp = shapefile.Writer(shapeType=shapefile.POINT)
        for name,dtype in par_info.dtypes.iteritems():
            if dtype == object:
                shp.field(name=name,fieldType='C',size=50)
            elif dtype in [int,np.int,np.int64,np.int32]:
                shp.field(name=name, fieldType='N', size=50, decimal=0)
            elif dtype in [float,np.float,np.float32,np.float64]:
                shp.field(name=name, fieldType='N', size=50, decimal=10)
            else:
                raise Exception("unrecognized field type in par_info:{0}:{1}".format(name,dtype))

        #some pandas awesomeness..
        par_info.apply(lambda x:shp.point(x.x,x.y), axis=1)
        par_info.apply(lambda x:shp.record(*x),axis=1)
        try:
            shp.save(shapename)
        except:
            shp.close()

        shp = shapefile.Reader(shapename)
        assert shp.numRecords == par_info.shape[0]
    return par_info


def pp_file_to_dataframe(pp_filename):

    """ read a pilot point file to a pandas Dataframe

    Parameters
    ----------
    pp_filename : str
        pilot point file

    Returns
    -------
    df : pandas.DataFrame
        a dataframe with pp_utils.PP_NAMES for columns

    """

    df = pd.read_csv(pp_filename, delim_whitespace=True,
                     header=None, names=PP_NAMES,usecols=[0,1,2,3,4])
    df.loc[:,"name"] = df.name.apply(str).apply(str.lower)
    return df

def pp_tpl_to_dataframe(tpl_filename):
    """ read a pilot points template file to a pandas dataframe

    Parameters
    ----------
    tpl_filename : str
        pilot points template file

    Returns
    -------
    df : pandas.DataFrame
        a dataframe with "parnme" included

    """
    inlines = open(tpl_filename, 'r').readlines()
    header = inlines.pop(0)
    marker = header.strip().split()[1]
    assert len(marker) == 1
    usecols = [0,1,2,3]
    df = pd.read_csv(tpl_filename, delim_whitespace=True,skiprows=1,
                     header=None, names=PP_NAMES[:-1],usecols=usecols)
    df.loc[:,"name"] = df.name.apply(str).apply(str.lower)
    df["parnme"] = [i.split(marker)[1].strip() for i in inlines]


    return df

def write_pp_shapfile(pp_df,shapename=None):
    """write pilot points dataframe to a shapefile

    Parameters
    ----------
    pp_df : pandas.DataFrame or str
        pilot point dataframe or a pilot point filename.  Dataframe
        must include "x" and "y"
    shapename : str
        shapefile name.  If None, pp_df must be str and shapefile
        is saved as <pp_df>.shp

    Note
    ----
        requires pyshp

    """
    try:
        import shapefile
    except Exception as e:
        raise Exception("error importing shapefile: {0}, \ntry pip install pyshp...".format(str(e)))

    if not isinstance(pp_df,list):
        pp_df = [pp_df]
    dfs = []
    for pp in pp_df:
        if isinstance(pp,pd.DataFrame):
            dfs.append(pp)
        elif isinstance(pp,str):
            dfs.append(pp_file_to_dataframe(pp))
        else:
            raise Exception("unsupported arg type:{0}".format(type(pp)))

    if shapename is None:
        shapename = "pp_locs.shp"
    try:
        shp = shapefile.Writer(shapeType=shapefile.POINT)
    except:
        shp = shapefile.Writer(target=shapename, shapeType=shapefile.POINT)
    for name, dtype in dfs[0].dtypes.iteritems():
        if dtype == object:
            shp.field(name=name, fieldType='C', size=50)
        elif dtype in [int, np.int, np.int64, np.int32]:
            shp.field(name=name, fieldType='N', size=50, decimal=0)
        elif dtype in [float, np.float, np.float32, np.float32]:
            shp.field(name=name, fieldType='N', size=50, decimal=8)
        else:
            raise Exception("unrecognized field type in pp_df:{0}:{1}".format(name, dtype))


    # some pandas awesomeness..
    for df in dfs:
        #df.apply(lambda x: shp.poly([[[x.x, x.y]]]), axis=1)
        df.apply(lambda x: shp.point(x.x, x.y), axis=1)
        df.apply(lambda x: shp.record(*x), axis=1)

    try:
        shp.save(shapename)
    except:
        shp.close()




def write_pp_file(filename,pp_df):
    """write a pilot points dataframe to a pilot points file

    Parameters
    ----------
    filename : str
        pilot points file to write
    pp_df : pandas.DataFrame
        a dataframe that has columns "x","y","zone", and "value"

    """
    with open(filename,'w') as f:
       f.write(pp_df.to_string(col_space=0,
                                columns=PP_NAMES,
                                formatters=PP_FMT,
                                justify="right",
                                header=False,
                                index=False) + '\n')


def pilot_points_to_tpl(pp_file,tpl_file=None,name_prefix=None):
    """write a template file for a pilot points file

    Parameters
    ----------
    pp_file : str
        pilot points file
    tpl_file : str
        template file name to write.  If None, append ".tpl" to
        the pp_file arg. Default is None
    name_prefix : str
        name to prepend to parameter names for each pilot point.  For example,
        if ``name_prefix = "hk_"``, then each pilot point parameter will be named
        "hk_0001","hk_0002", etc.  If None, parameter names from pp_df.name
        are used.  Default is None.

    Returns
    -------
        pp_df : pandas.DataFrame
            a dataframe with pilot point information (name,x,y,zone,parval1)
             with the parameter information (parnme,tpl_str)
    """

    if isinstance(pp_file,pd.DataFrame):
        pp_df = pp_file
        assert tpl_file is not None
    else:
        assert os.path.exists(pp_file)
        pp_df = pd.read_csv(pp_file, delim_whitespace=True,
                            header=None, names=PP_NAMES)

    if tpl_file is None:
        tpl_file = pp_file + ".tpl"

    if name_prefix is not None:
        digits = str(len(str(pp_df.shape[0])))
        fmt = "{0:0"+digits+"d}"
        names = [name_prefix+fmt.format(i) for i in range(pp_df.shape[0])]
    else:
        names = pp_df.name.copy()

    too_long = []
    for name in names:
        if len(name) > 12:
            too_long.append(name)
    if len(too_long) > 0:
        raise Exception("the following parameter names are too long:" +\
                        ",".join(too_long))

    tpl_entries = ["~    {0}    ~".format(name) for name in names]
    pp_df.loc[:,"tpl"] = tpl_entries
    pp_df.loc[:,"parnme"] = names


    f_tpl = open(tpl_file,'w')
    f_tpl.write("ptf ~\n")
    f_tpl.write(pp_df.to_string(col_space=0,
                              columns=["name","x","y","zone","tpl"],
                              formatters=PP_FMT,
                              justify="left",
                              header=False,
                              index=False) + '\n')

    return pp_df



"""module for FOSM-based uncertainty analysis using a
linearized form of Bayes equation known as the Schur compliment
"""

from __future__ import print_function, division
import numpy as np
import pandas as pd
from pyemu.la import LinearAnalysis
from pyemu.mat import Cov, Matrix

class Schur(LinearAnalysis):
    """derived type for prior and posterior uncertainty and data-worth
    analysis using Schur compliment

    Parameters
    ----------
    **kwargs : dict
        keyword arguments to pass to the LinearAnalysis constructor.  See
        LinearAnalysis definition for argument types

    Note
    ----
    Same call signature as the base LinearAnalysis class

    Example
    -------
    ``>>>import pyemu``

    ``>>>sc = pyemu.Schur(jco="pest.jcb")``
    
    """
    def __init__(self,jco,**kwargs):
        self.__posterior_prediction = None
        self.__posterior_parameter = None
        super(Schur,self).__init__(jco,**kwargs)


    @property
    def pandas(self):
        """get a pandas dataframe of prior and posterior for all predictions

        Returns:
            pandas.DataFrame : pandas.DataFrame
                a dataframe with prior and posterior uncertainty estimates
                for all forecasts (predictions)
        """
        names,prior,posterior = [],[],[]
        for iname,name in enumerate(self.posterior_parameter.row_names):
            names.append(name)
            posterior.append(np.sqrt(float(
                self.posterior_parameter[iname, iname]. x)))
            iprior = self.parcov.row_names.index(name)
            prior.append(np.sqrt(float(self.parcov[iprior, iprior].x)))
        for pred_name, pred_var in self.posterior_prediction.items():
            names.append(pred_name)
            posterior.append(np.sqrt(pred_var))
            prior.append(self.prior_prediction[pred_name])
        return pd.DataFrame({"posterior": posterior, "prior": prior},
                                index=names)

    @property
    def posterior_parameter(self):
        """get the posterior parameter covariance matrix.  If Schur.__posterior_parameter
        is None, the posterior parameter covariance matrix is calculated via
        Schur compliment before returning

        Returns
        -------
        posterior_parameter : pyemu.Cov
            the posterior parameter covariance matrix

        Note
        ----
        returns a reference

        Example
        -------
        ``>>>import pyemu``

        ``>>>sc = pyemu.Schur(jco="pest.jcb")``

        ``>>>post_cov = sc.posterior_parameter``

        ``>>>post_cov.to_ascii("post.cov")``


        """
        if self.__posterior_parameter is not None:
            return self.__posterior_parameter
        else:
            self.clean()
            self.log("Schur's complement")
            try:
                pinv = self.parcov.inv
                r = self.xtqx + pinv
                r = r.inv
            except Exception as e:
                self.xtqx.to_binary("xtqx.err.jcb")
                pinv.to_ascii("parcov_inv.err.cov")
                self.logger.warn("error forming schur's complement: {0}".
                                format(str(e)))
                self.logger.warn("problemtic xtqx saved to xtqx.err.jcb")
                self.logger.warn("problematic inverse parcov saved to parcov_inv.err.cov")
                raise Exception("error forming schur's complement: {0}".
                                format(str(e)))
            assert r.row_names == r.col_names
            self.__posterior_parameter = Cov(r.x, row_names=r.row_names,
                                             col_names=r.col_names)
            self.log("Schur's complement")
            return self.__posterior_parameter

    @property
    def map_parameter_estimate(self):
        """ get the posterior expectation for parameters using Bayes linear
        estimation

        Returns
        -------
        post_expt : pandas.DataFrame
            a dataframe with prior and posterior parameter expectations

        """
        res = self.pst.res
        assert res is not None
        # build the prior expectation parameter vector
        prior_expt = self.pst.parameter_data.loc[:,["parval1"]].copy()
        islog = self.pst.parameter_data.partrans == "log"
        prior_expt.loc[islog] = prior_expt.loc[islog].apply(np.log10)
        prior_expt = Matrix.from_dataframe(prior_expt)
        prior_expt.col_names = ["prior_expt"]
        # build the residual vector
        res_vec = Matrix.from_dataframe(res.loc[:,["residual"]])

        # form the terms of Schur's complement
        b = self.parcov * self.jco.T
        c = ((self.jco * self.parcov * self.jco.T) + self.obscov).inv
        bc = Matrix((b * c).x, row_names=b.row_names, col_names=c.col_names)

        # calc posterior expectation
        upgrade = bc * res_vec
        upgrade.col_names = ["prior_expt"]
        post_expt = prior_expt + upgrade

        # post processing - back log transform
        post_expt = pd.DataFrame(data=post_expt.x,index=post_expt.row_names,
                                 columns=["post_expt"])
        post_expt.loc[:,"prior_expt"] = prior_expt.x.flatten()
        post_expt.loc[islog,:] = 10.0**post_expt.loc[islog,:]
        # unc_sum = self.get_parameter_summary()
        # post_expt.loc[:,"standard_deviation"] = unc_sum.post_var.apply(np.sqrt)
        post_expt.sort_index(inplace=True)
        return post_expt

    @property
    def map_forecast_estimate(self):
        """ get the prior and posterior forecast (prediction) expectations.

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            dataframe with prior and posterior forecast expected values

        """
        assert self.forecasts is not None
        islog = self.pst.parameter_data.partrans == "log"
        par_map = self.map_parameter_estimate
        par_map.loc[islog,:] = np.log10(par_map.loc[islog,:])
        par_map = Matrix.from_dataframe(par_map.loc[:,["post_expt"]])
        posts,priors = [],[]
        post_expt = (self.predictions.T * par_map).to_dataframe()
        for fname in self.forecast_names:
            #fname = forecast.col_names[0]
            pr = self.pst.res.loc[fname,"modelled"]
            priors.append(pr)
            posts.append(pr + post_expt.loc[fname,"post_expt"])
        return pd.DataFrame(data=np.array([priors,posts]).transpose(),
                            columns=["prior_expt","post_expt"],
                            index=self.forecast_names)

    @property
    def posterior_forecast(self):
        """thin wrapper around posterior_prediction
        """
        return self.posterior_prediction

    @property
    def posterior_prediction(self):
        """get posterior forecast (prediction) variances

        Returns
        -------
        dict : dict
            a dictionary of forecast names, posterior variance pairs

        Note
        ----
        This method is not as easy to use as Schur.get_forecast_summary(), please
        use it instead

        """
        if self.__posterior_prediction is not None:
            return self.__posterior_prediction
        else:
            if self.predictions is not None:
                self.log("propagating posterior to predictions")

                post_cov = self.predictions.T *\
                            self.posterior_parameter * self.predictions
                self.__posterior_prediction = {n:v for n,v in
                                          zip(post_cov.row_names,
                                              np.diag(post_cov.x))}
                self.log("propagating posterior to predictions")
            else:
                self.__posterior_prediction = {}
            return self.__posterior_prediction

    def get_parameter_summary(self,include_map=False):
        """get a summary of the parameter uncertainty

        Parameters
        ----------
        include_map : bool
            if True, add the prior and posterior expectations
            and report standard deviation instead of variance

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            dataframe of prior,posterior variances and percent
            uncertainty reduction of each parameter

        Note
        ----
        this is the primary method for accessing parameter uncertainty
        estimates - use this!

        Example
        -------
        ``>>>import matplotlib.pyplot as plt``

        ``>>>import pyemu``

        ``>>>sc = pyemu.Schur(jco="pest.jcb")``

        ``>>>sc = pyemu.Schur(jco="pest.jcb",forecasts=["fore1","fore2"])``

        ``>>>par_sum = sc.get_parameter_summary()``

        ``>>>par_sum.plot(kind="bar")``

        ``>>>plt.show()``

        """
        prior_mat = self.parcov.get(self.posterior_parameter.col_names)
        if prior_mat.isdiagonal:
            prior = prior_mat.x.flatten()
        else:
            prior = np.diag(prior_mat.x)
        post = np.diag(self.posterior_parameter.x)
        if include_map:
            par_data = self.map_parameter_estimate
            prior = pd.DataFrame(data=prior,index=prior_mat.col_names)
            islog = self.pst.parameter_data.partrans == "log"
            par_data.loc[islog,:] = np.log10(par_data.loc[islog,:])
            par_data.loc[:,"prior_stdev"] = prior
            post = pd.DataFrame(data=post,index=prior.index)
            par_data.loc[:,"post_stdev"] = post
            par_data.loc[:,"is_log"] = islog
            return par_data
        else:
            ureduce = 100.0 * (1.0 - (post / prior))

            return pd.DataFrame({"prior_var":prior,"post_var":post,
                                     "percent_reduction":ureduce},
                                    index=self.posterior_parameter.col_names)

    def get_forecast_summary(self, include_map=False):
        """get a summary of the forecast uncertainty

        Parameters
        ----------
        include_map : bool
            if True, add the prior and posterior expectations
            and report standard deviation instead of variance

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
                dataframe of prior,posterior variances and percent
                uncertainty reduction of each parameter

        Note
        ----
        this is the primary method for accessing forecast uncertainty
        estimates - use this!

        Example
        -------
        ``>>>import matplotlib.pyplot as plt``

        ``>>>import pyemu``

        This usage assumes you have set the ``++forecasts()`` argument in the
        control file:

        ``>>>sc = pyemu.Schur(jco="pest.jcb")``

        or, you can pass the forecasts directly, assuming the forecasts are
        names of zero-weight observations:

        ``>>>sc = pyemu.Schur(jco="pest.jcb",forecasts=["fore1","fore2"])``

        ``>>>fore_sum = sc.get_forecast_summary()``

        ``>>>fore_sum.plot(kind="bar")``

        ``>>>plt.show()``

        """
        sum = {"prior_var":[], "post_var":[], "percent_reduction":[]}
        for forecast in self.prior_forecast.keys():
            pr = self.prior_forecast[forecast]
            pt = self.posterior_forecast[forecast]
            ur = 100.0 * (1.0 - (pt/pr))
            sum["prior_var"].append(pr)
            sum["post_var"].append(pt)
            sum["percent_reduction"].append(ur)
        df = pd.DataFrame(sum,index=self.prior_forecast.keys())

        if include_map:
            df.loc[:,"prior_stdev"] = df.pop("prior_var").apply(np.sqrt)
            df.loc[:,"post_stdev"] = df.pop("post_var").apply(np.sqrt)
            df.pop("percent_reduction")
            forecast_map = self.map_forecast_estimate
            df.loc[:,"prior_expt"] = forecast_map.prior_expt
            df.loc[:,"post_expt"] = forecast_map.post_expt
            return df
        return pd.DataFrame(sum,index=self.prior_forecast.keys())

    def __contribution_from_parameters(self, parameter_names):
        """private method get the prior and posterior uncertainty reduction as a result of
        some parameter becoming perfectly known

        Parameters
        ----------
        parameter_names : list
            parameter that are perfectly known

        Returns
        -------
        dict : dict
            dictionary of forecast name,  [prior uncertainty w/o parameter_names,
                % posterior uncertainty w/o parameter names]

        Note
        ----
        this method is used by get_parameter_contribution() method - don't
        call this method directly

        """


        #get the prior and posterior for the base case
        bprior,bpost = self.prior_prediction, self.posterior_prediction
        #get the prior and posterior for the conditioned case
        la_cond = self.get_conditional_instance(parameter_names)
        cprior,cpost = la_cond.prior_prediction, la_cond.posterior_prediction
        return cprior,cpost

    def get_conditional_instance(self, parameter_names):
        """ get a new Schur instance that includes conditional update from
        some parameters becoming known perfectly

        Parameters
        ----------
        parameter_names : list
            parameters that are to be treated as notionally perfectly
            known

        Returns
        -------
        la_cond : Schur
            a new Schur instance conditional on perfect knowledge
            of some parameters

        Note
        ----
        this method is used by the get_parameter_contribution() method -
        don't call this method directly

        """
        if not isinstance(parameter_names, list):
            parameter_names = [parameter_names]

        for iname, name in enumerate(parameter_names):
            name = str(name).lower()
            parameter_names[iname] = name
            assert name in self.jco.col_names,\
                "contribution parameter " + name + " not found jco"
        keep_names = []
        for name in self.jco.col_names:
            if name not in parameter_names:
                keep_names.append(name)
        if len(keep_names) == 0:
            raise Exception("Schur.contribution_from_Parameters " +
                            "atleast one parameter must remain uncertain")
        #get the reduced predictions
        if self.predictions is None:
            raise Exception("Schur.contribution_from_Parameters " +
                            "no predictions have been set")
        # cond_preds = []
        # for pred in self.predictions:
        #     cond_preds.append(pred.get(keep_names, pred.col_names))
        cond_preds = self.predictions.get(row_names=keep_names)
        la_cond = Schur(jco=self.jco.get(self.jco.row_names, keep_names),
                        parcov=self.parcov.condition_on(parameter_names),
                        obscov=self.obscov, predictions=cond_preds,verbose=False)
        return la_cond

    def get_par_contribution(self,parlist_dict=None,include_prior_results=False):
        """get a dataframe the prior and posterior uncertainty
        reduction as a result of some parameter becoming perfectly known

        Parameters
        ----------
        parlist_dict : dict
            a nested dictionary-list of groups of parameters
            that are to be treated as perfectly known.  key values become
            row labels in returned dataframe.  If None, each adjustable parameter
            is sequentially treated as known and the returned dataframe
            has row labels for each adjustable parameter
        include_prior_results : bool
            flag to return a multi-indexed dataframe with both conditional
            prior and posterior forecast uncertainty estimates.  Default is False

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            a dataframe that summarizes the parameter contribution analysis.
            The dataframe has index (row labels) of the keys in parlist_dict
            and a column labels of forecast names.  The values in the dataframe
            are the posterior variance of the forecast conditional on perfect
            knowledge of the parameters in the values of parlist_dict.  Varies
            depending on `include_prior_results`.

        Example
        -------
        ``>>>import pyemu``

        ``>>>sc = pyemu.Schur(jco="pest.jcb")``

        ``>>>df = sc.get_par_contribution()``

        """
        self.log("calculating contribution from parameters")
        if parlist_dict is None:
            parlist_dict = {}#dict(zip(self.pst.adj_par_names,self.pst.adj_par_names))
            # make sure all of the adjustable pars are in the jco
            for pname in self.pst.adj_par_names:
                if pname in self.jco.col_names:
                    parlist_dict[pname] = pname
        else:
            if type(parlist_dict) == list:
                parlist_dict = dict(zip(parlist_dict,parlist_dict))

        results = {}
        names = ["base"]
        for forecast in self.prior_forecast.keys():
            pr = self.prior_forecast[forecast]
            pt = self.posterior_forecast[forecast]
            #reduce = 100.0 * ((pr - pt) / pr)
            results[(forecast,"prior")] = [pr]
            results[(forecast,"post")] = [pt]
            #results[(forecast,"percent_reduce")] = [reduce]
        for case_name,par_list in parlist_dict.items():
            if len(par_list) == 0:
                continue
            names.append(case_name)
            self.log("calculating contribution from: " + str(par_list))
            case_prior,case_post = self.__contribution_from_parameters(par_list)
            self.log("calculating contribution from: " + str(par_list))
            for forecast in case_prior.keys():
                pr = case_prior[forecast]
                pt = case_post[forecast]
                #reduce = 100.0 * ((pr - pt) / pr)
                results[(forecast, "prior")].append(pr)
                results[(forecast, "post")].append(pt)
                #results[(forecast, "percent_reduce")].append(reduce)

        df = pd.DataFrame(results,index=names)
        #base = df.loc["base",df.columns.get_level_values(1)=="post"]
        #df = 1.0 - (df.loc[:,df.columns.get_level_values(1)=="post"] / base)

        self.log("calculating contribution from parameters")
        if include_prior_results:
            return df
        else:
            df = df.xs("post", level=1, drop_level=True, axis=1)
            return df

    def get_par_group_contribution(self, include_prior_results=False):
        """get the forecast uncertainty contribution from each parameter
        group.  Just some sugar for get_contribution_dataframe() - this method
        automatically constructs the parlist_dict argument where the keys are the
        group names and the values are the adjustable parameters in the groups

        Parameters
        ----------
        include_prior_results : bool
            flag to return a multi-indexed dataframe with both conditional
            prior and posterior forecast uncertainty estimates.  Default is False


        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            a dataframe that summarizes the parameter contribution analysis.
            The dataframe has index (row labels) that are the parameter groups
            and a column labels of forecast names.  The values in the dataframe
            are the posterior variance of the forecast conditional on perfect
            knowledge of the adjustable parameters in each parameter groups
            Varies depending on `include_prior_results`.

        """
        pargrp_dict = {}
        par = self.pst.parameter_data
        groups = par.groupby("pargp").groups
        for grp,idxs in groups.items():
            #pargrp_dict[grp] = list(par.loc[idxs,"parnme"])
            pargrp_dict[grp] = [pname for pname in list(par.loc[idxs,"parnme"])
                                if pname in self.jco.col_names and pname in self.parcov.row_names]
        return self.get_par_contribution(pargrp_dict,include_prior_results=include_prior_results)

    def get_added_obs_importance(self,obslist_dict=None,base_obslist=None,
                                 reset_zero_weight=False):
        """get a dataframe fo the posterior uncertainty
        as a results of added some observations

        Parameters
        ----------
        obslist_dict : dict
            a nested dictionary-list of groups of observations
            that are to be treated as gained.  key values become
            row labels in returned dataframe. If None, then every zero-weighted
            observation is tested sequentially. Default is None
        base_obslist : list
            observation names to treat as the "existing" observations.
            The values of obslist_dict will be added to this list during
            each test.  If None, then the values in obslist_dict will
            be treated as the entire calibration dataset.  That is, there
            are no existing data. Default is None.  Standard practice would
            be to pass this argument as Schur.pst.nnz_obs_names.
        reset_zero_weight : (boolean or float)
            a flag to reset observations with zero weight in either
            obslist_dict or base_obslist. The value of reset_zero_weights
            can be cast to a float,then that value will be assigned to
            zero weight obs.  Otherwise, zero weight obs will be given a
            weight of 1.0.  Default is False.

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            dataframe with row labels (index) of obslist_dict.keys() and
            columns of forecast_name.  The values in the dataframe are the
            posterior variance of the forecasts resulting from notional inversion
            using the observations in obslist_dict[key value] plus the observations
            in base_obslist (if any)

        Note
        ----
        all observations listed in obslist_dict and base_obslist with zero
        weights will be dropped unless reset_zero_weight is set

        Example
        -------
        ``>>>import pyemu``

        ``>>>sc = pyemu.Schur(jco="pest.jcb")``

        ``>>>df = sc.get_added_obs_importance(base_obslist=sc.pst.nnz_obs_names,reset_zero=True)``


        """

        if obslist_dict is not None:
            if type(obslist_dict) == list:
                obslist_dict = dict(zip(obslist_dict,obslist_dict))

        reset = False
        if reset_zero_weight is not False:
            if not self.obscov.isdiagonal:
                raise NotImplementedError("cannot reset weights for non-"+\
                                          "diagonal obscov")
            reset = True
            try:
                weight = float(reset_zero_weight)
            except:
                weight = 1.0
            self.logger.statement("resetting zero weights to {0}".format(weight))
            # make copies of the original obscov and pst
            #org_obscov = self.obscov.get(self.obscov.row_names)
            org_obscov = self.obscov.copy()
            org_pst = self.pst.get()

        obs = self.pst.observation_data
        obs.index = obs.obsnme

        # if we don't care about grouping obs, then just reset all weights at once
        if base_obslist is None and obslist_dict is None and reset:
            onames = [name for name in self.pst.zero_weight_obs_names
                      if name in self.jco.obs_names and name in self.obscov.row_names]
            obs.loc[onames,"weight"] = weight

        # if needed reset the zero-weight obs in base_obslist
        if base_obslist is not None and reset:
            # check for zero
            self.log("resetting zero weight obs in base_obslist")
            self.pst._adjust_weights_by_list(base_obslist, weight)
            self.log("resetting zero weight obs in base_obslist")

        if base_obslist is None:
            base_obslist = []
        else:
            if type(base_obslist) != list:
                self.logger.lraise("Schur.get_added_obs)_importance: base_obslist must be" +
                                   " type 'list', not {0}".format(str(type(base_obslist))))
        # if needed reset the zero-weight obs in obslist_dict
        if obslist_dict is not None and reset:
            z_obs = []
            for case,obslist in obslist_dict.items():
                if not isinstance(obslist,list):
                    obslist_dict[case] = [obslist]
                    obslist = [obslist]
                inboth = set(base_obslist).intersection(set(obslist))
                if len(inboth) > 0:
                    raise Exception("observation(s) listed in both "+\
                                    "base_obslist and obslist_dict: "+\
                                    ','.join(inboth))
                z_obs.extend(obslist)
            self.log("resetting zero weight obs in obslist_dict")
            self.pst._adjust_weights_by_list(z_obs, weight)
            self.log("resetting zero weight obs in obslist_dict")

        # for a comprehensive obslist_dict
        if obslist_dict is None and reset:
            obs = self.pst.observation_data
            obs.index = obs.obsnme
            onames = [name for name in self.pst.zero_weight_obs_names
                      if name in self.jco.obs_names and name in self.obscov.row_names]
            obs.loc[onames,"weight"] = weight

        if obslist_dict is None:
            obslist_dict = {name:name for name in self.pst.nnz_obs_names if name\
                            in self.jco.obs_names and name in self.obscov.row_names}

        # reset the obs cov from the newly adjusted weights
        if reset:
            self.log("resetting self.obscov")
            self.reset_obscov(self.pst)
            self.log("resetting self.obscov")

        results = {}
        names = ["base"]

        if base_obslist is None or len(base_obslist) == 0:
            self.logger.statement("no base observation passed, 'base' case"+
                                  " is just the prior of the forecasts")
            for forecast,pr in self.prior_forecast.items():
                results[forecast] = [pr]
            # reset base obslist for use later
            base_obslist = []

        else:
            base_posterior = self.get(par_names=self.jco.par_names,
                                      obs_names=base_obslist).posterior_forecast
            for forecast,pt in base_posterior.items():
                results[forecast] = [pt]

        for case_name,obslist in obslist_dict.items():
            names.append(case_name)
            if not isinstance(obslist,list):
                obslist = [obslist]
            self.log("calculating importance of observations by adding: " +
                     str(obslist) + '\n')
            # this case is the combination of the base obs plus whatever unique
            # obs names in obslist
            case_obslist = list(base_obslist)
            dedup_obslist = [oname for oname in obslist if oname not in case_obslist]
            case_obslist.extend(dedup_obslist)
            #print(self.pst.observation_data.loc[case_obslist,:])
            case_post = self.get(par_names=self.jco.par_names,
                                 obs_names=case_obslist).posterior_forecast
            for forecast,pt in case_post.items():
                results[forecast].append(pt)
            self.log("calculating importance of observations by adding: " +
                     str(obslist) + '\n')
        df = pd.DataFrame(results,index=names)


        if reset:
            self.reset_obscov(org_obscov)
            self.reset_pst(org_pst)

        return df

    def get_removed_obs_importance(self,obslist_dict=None,
                                   reset_zero_weight=False):
        """get a dataframe the posterior uncertainty
        as a result of losing some observations

        Parameters
        ----------
        obslist_dict : dict
            dictionary of groups of observations
            that are to be treated as lost.  key values become
            row labels in returned dataframe. If None, then test every
            (nonzero weight - see reset_zero_weight) observation
        reset_zero_weight : bool or float
            a flag to reset observations with zero weight in obslist_dict.
            If the value of reset_zero_weights can be cast to a float,
            then that value will be assigned to zero weight obs.  Otherwise,
            zero weight obs will be given a weight of 1.0

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            a dataframe with index of obslist_dict.keys() and columns
            of forecast names.  The values in the dataframe are the posterior
            variances of the forecasts resulting from losing the information
            contained in obslist_dict[key value]

        Note
        ----
        all observations listed in obslist_dict with zero
        weights will be dropped unless reset_zero_weight is set

        Example
        -------
        ``>>>import pyemu``

        ``>>>sc = pyemu.Schur(jco="pest.jcb")``

        ``df = sc.get_removed_obs_importance()``

        """


        if obslist_dict is not None:
            if type(obslist_dict) == list:
                obslist_dict = dict(zip(obslist_dict,obslist_dict))

        elif reset_zero_weight is False and self.pst.nnz_obs == 0:
            raise Exception("not resetting weights and there are no non-zero weight obs to remove")

        reset = False
        if reset_zero_weight is not False:
            if not self.obscov.isdiagonal:
                raise NotImplementedError("cannot reset weights for non-"+\
                                          "diagonal obscov")
            reset = True
            try:
                weight = float(reset_zero_weight)
            except:
                weight = 1.0
            self.logger.statement("resetting zero weights to {0}".format(weight))
            # make copies of the original obscov and pst
            org_obscov = self.obscov.get(self.obscov.row_names)
            org_pst = self.pst.get()

        self.log("calculating importance of observations")
        if reset and obslist_dict is None:
            obs = self.pst.observation_data
            onames = [name for name in self.pst.zero_weight_obs_names
                      if name in self.jco.obs_names and name in self.obscov.row_names]
            obs.loc[onames,"weight"] = weight

        if obslist_dict is None:
            obslist_dict = dict(zip(self.pst.nnz_obs_names,
                                        self.pst.nnz_obs_names))


        elif reset:
            self.pst.observation_data.index = self.pst.observation_data.obsnme
            for name,obslist in obslist_dict.items():
                self.log("resetting weights in obs in group {0}".format(name))
                self.pst._adjust_weights_by_list(obslist,weight)
                self.log("resetting weights in obs in group {0}".format(name))

        for case,obslist in obslist_dict.items():
            if not isinstance(obslist,list):
                obslist = [obslist]
            obslist_dict[case] = obslist


        if reset:
            self.log("resetting self.obscov")
            self.reset_obscov(self.pst)
            self.log("resetting self.obscov")

        results = {}
        names = ["base"]
        for forecast,pt in self.posterior_forecast.items():
            results[forecast] = [pt]
        for case_name,obslist in obslist_dict.items():
            if not isinstance(obslist,list):
                obslist = [obslist]
            names.append(case_name)
            self.log("calculating importance of observations by removing: " +
                     str(obslist) + '\n')
            # check for missing names
            missing_onames = [oname for oname in obslist if oname not in self.jco.obs_names]
            if len(missing_onames) > 0:
                raise Exception("case {0} has observation names ".format(case_name) + \
                                "not found: " + ','.join(missing_onames))
            # find the set difference between obslist and jco obs names
            #diff_onames = [oname for oname in self.jco.obs_names if oname not in obslist]
            diff_onames = [oname for oname in self.nnz_obs_names if oname not in obslist and oname not in self.forecast_names]


            # calculate the increase in forecast variance by not using the obs
            # in obslist
            case_post = self.get(par_names=self.jco.par_names,
                                 obs_names=diff_onames).posterior_forecast

            for forecast,pt in case_post.items():
                results[forecast].append(pt)
        df = pd.DataFrame(results,index=names)
        self.log("calculating importance of observations by removing: " +
                     str(obslist) + '\n')

        if reset:
            self.reset_obscov(org_obscov)
            self.reset_pst(org_pst)
        return df

    def obs_group_importance(self):
        obsgrp_dict = {}
        obs = self.pst.observation_data
        obs.index = obs.obsnme
        obs = obs.loc[self.jco.row_names,:]
        groups = obs.groupby("obgnme").groups
        for grp, idxs in groups.items():
            obsgrp_dict[grp] = list(obs.loc[idxs,"obsnme"])
        return obsgrp_dict
        
    def get_removed_obs_group_importance(self):
        return self.get_removed_obs_importance(self.obs_group_importance())

    def get_added_obs_group_importance(self):
        return self.get_added_obs_importance(self.obs_group_importance())

    def next_most_important_added_obs(self,forecast=None,niter=3, obslist_dict=None,
                                      base_obslist=None,
                                      reset_zero_weight=False):
        """find the most important observation(s) by sequentially evaluating
        the importance of the observations in obslist_dict. The most important observations
        from each iteration is added to base_obslist and removed obslist_dict for the
        next iteration.  In this way, the added observation importance values include
        the conditional information from the last iteration.

        Parameters
        ----------
        forecast : str
            name of the forecast to use in the ranking process.  If
            more than one forecast has been listed, this argument is required
        niter : int
            number of sequential iterations
        obslist_dict dict:
            nested dictionary-list of  groups of observations
            that are to be treated as gained.  key values become
            row labels in result dataframe. If None, then test every observation
            individually
        base_obslist : list
            observation names to treat as the "existing" observations.
            The values of obslist_dict will be added to this list during testing.
            If None, then each list in the values of obslist_dict will be
            treated as an individual calibration dataset.
        reset_zero_weight : (boolean or float)
            a flag to reset observations with zero weight in either
            obslist_dict or base_obslist. If the value of reset_zero_weights
            can be cast to a float,then that value will be assigned to
            zero weight obs.  Otherwise, zero weight obs will be given a weight of 1.0

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            DataFrame with columns of best obslist_dict key for each iteration.
            Columns of forecast variance percent reduction for this iteration,
            (percent reduction compared to initial base case)


        Example
        -------
        ``>>>import pyemu``

        ``>>>sc = pyemu.Schur(jco="pest.jcb")``

        ``>>>df = sc.next_most_added_importance_obs(forecast="fore1",``

        ``>>>      base_obslist=sc.pst.nnz_obs_names,reset_zero=True``
        """


        if forecast is None:
            assert self.forecasts.shape[1] == 1,"forecast arg list one and only one" +\
                                            " forecast"
            forecast = self.forecasts[0].col_names[0]
        #elif forecast not in self.prediction_arg:
        #    raise Exception("forecast {0} not found".format(forecast))

        else:
            forecast = forecast.lower()
            found = False
            for fore in self.forecasts.col_names:
                if fore == forecast:
                    found = True
                    break
            if not found:
                raise Exception("forecast {0} not found".format(forecast))




        if base_obslist:
            obs_being_used = list(base_obslist)
        else:
            obs_being_used = []

        best_case, best_results = [],[]
        for iiter in range(niter):
            self.log("next most important added obs iteration {0}".format(iiter+1))
            df = self.get_added_obs_importance(obslist_dict=obslist_dict,
                                                   base_obslist=obs_being_used,
                                                   reset_zero_weight=reset_zero_weight)

            if iiter == 0:
                init_base = df.loc["base",forecast].copy()
            fore_df = df.loc[:,forecast]
            fore_diff_df = fore_df - fore_df.loc["base"]
            fore_diff_df.sort_values(inplace=True)
            iter_best_name = fore_diff_df.index[0]
            iter_best_result = df.loc[iter_best_name,forecast]
            iter_base_result = df.loc["base",forecast]
            diff_percent_init = 100.0 * (init_base -
                                              iter_best_result) / init_base
            diff_percent_iter = 100.0 * (iter_base_result -
                                              iter_best_result) / iter_base_result
            self.log("next most important added obs iteration {0}".format(iiter+1))


            best_results.append([iter_best_name,iter_best_result,
                                 diff_percent_iter,diff_percent_init])
            best_case.append(iter_best_name)

            if iter_best_name.lower() == "base":
                break

            if obslist_dict is None:
                onames = [iter_best_name]
            else:
                onames = obslist_dict.pop(iter_best_name)
            if not isinstance(onames,list):
                onames = [onames]
            obs_being_used.extend(onames)
        columns = ["best_obs",forecast+"_variance",
                   "unc_reduce_iter_base","unc_reduce_initial_base"]
        return pd.DataFrame(best_results,index=best_case,columns=columns)

    def next_most_par_contribution(self,niter=3,forecast=None,parlist_dict=None):
        """find the largest parameter(s) contribution for prior and posterior
        forecast  by sequentially evaluating the contribution of parameters in
        parlist_dict.  The largest contributing parameters from each iteration are
        treated as known perfectly for the remaining iterations.  In this way, the
        next iteration seeks the next most influential group of parameters.

        Parameters
        ----------
        forecast : str
            name of the forecast to use in the ranking process.  If
            more than one forecast has been listed, this argument is required
        parlist_dict : dict
            a nested dictionary-list of groups of parameters
            that are to be treated as perfectly known.  key values become
            row labels in dataframe

        Returns
        -------
        pandas.DataFrame : pandas.DataFrame
            a dataframe with index of iteration number and columns
            of parlist_dict keys.  The values are the results of the knowing
            each parlist_dict entry expressed as posterior variance reduction

        """
        if forecast is None:
            assert len(self.forecasts) == 1,"forecast arg list one and only one" +\
                                            " forecast"
        elif forecast not in self.prediction_arg:
            raise Exception("forecast {0} not found".format(forecast))
        org_parcov = self.parcov.get(row_names=self.parcov.row_names)
        if parlist_dict is None:
            parlist_dict = dict(zip(self.pst.adj_par_names,self.pst.adj_par_names))

        base_prior,base_post = self.prior_forecast,self.posterior_forecast
        iter_results = [base_post[forecast].copy()]
        iter_names = ["base"]
        for iiter in range(niter):
            iter_contrib = {forecast:[base_post[forecast]]}
            iter_case_names = ["base"]
            self.log("next most par iteration {0}".format(iiter+1))

            for case,parlist in parlist_dict.items():
                iter_case_names.append(case)
                la_cond = self.get_conditional_instance(parlist)
                iter_contrib[forecast].append(la_cond.posterior_forecast[forecast])
            df = pd.DataFrame(iter_contrib,index=iter_case_names)
            df.sort_values(by=forecast,inplace=True)
            iter_best = df.index[0]
            self.logger.statement("next best iter {0}: {1}".format(iiter+1,iter_best))
            self.log("next most par iteration {0}".format(iiter+1))
            if iter_best.lower() == "base":
                break
            iter_results.append(df.loc[iter_best,forecast])
            iter_names.append(iter_best)
            self.reset_parcov(self.parcov.condition_on(parlist_dict.pop(iter_best)))

        self.reset_parcov(org_parcov)
        return pd.DataFrame(iter_results,index=iter_names)






"""This module contains classes for handling matrices in a linear algebra setting.
The primary objects are the Matrix() and Cov().  These objects overload most numerical
operators to autoalign the elements based on row and column names."""

from .mat_handler import Matrix, Cov, Jco, SparseMatrix, concat, save_coo

